@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2012 Con Kolivas
+   Copyright (C) 2006-2013 Con Kolivas
    Copyright (C) 1998 Andrew Tridgell
 
    Modified to use flat hash, memory limit and variable hash culling
@@ -104,7 +104,7 @@ struct rzip_state {
 	struct level *level;
 	tag hash_index[256];
 	struct hash_entry *hash_table;
-	i64 hash_bits;
+	char hash_bits;
 	i64 hash_count;
 	i64 hash_limit;
 	tag minimum_tag_mask;
@@ -115,7 +115,7 @@ struct rzip_state {
 	char chunk_bytes;
 	uint32_t cksum;
 	int fd_in, fd_out;
-	int stdin_eof;
+	char stdin_eof;
 	struct {
 		i64 inserts;
 		i64 literals;
@@ -173,14 +173,18 @@ static inline bool remap_high_sb(rzip_control *control, struct sliding_buffer *s
 static uchar *sliding_get_sb(rzip_control *control, i64 p)
 {
 	struct sliding_buffer *sb = &control->sb;
+	i64 sbo;
 
-	if (p >= sb->offset_low && p < sb->offset_low + sb->size_low)
-		return (sb->buf_low + p - sb->offset_low);
-	if (p >= sb->offset_high && p < (sb->offset_high + sb->size_high))
-		return (sb->buf_high + (p - sb->offset_high));
+	sbo = sb->offset_low;
+	if (p >= sbo && p < sbo + sb->size_low)
+		return (sb->buf_low + p - sbo);
+	sbo = sb->offset_high;
+	if (p >= sbo && p < (sbo + sb->size_high))
+		return (sb->buf_high + (p - sbo));
 	/* p is not within the low or high buffer range */
 	if (unlikely(!remap_high_sb(control, &control->sb, p)))
 		return NULL;
+	/* Use sb->offset_high directly since it will have changed */
 	return (sb->buf_high + (p - sb->offset_high));
 }
 
@@ -189,6 +193,26 @@ static uchar *single_get_sb(__maybe_unused rzip_control *control, i64 p)
 	return (control->sb.buf_low + p);
 }
 
+/* The length of continous range of the sliding buffer,
+ * starting from the offset P.
+ */
+static inline i64 sliding_get_sb_range(rzip_control *control, i64 p)
+{
+	struct sliding_buffer *sb = &control->sb;
+	i64 sbo, sbs;
+
+	sbo = sb->offset_low;
+	sbs = sb->size_low;
+	if (p >= sbo && p < sbo + sbs)
+		return (sbs - (p - sbo));
+	sbo = sb->offset_high;
+	sbs = sb->size_high;
+	if (likely(p >= sbo && p < (sbo + sbs)))
+		return (sbs - (p - sbo));
+
+	fatal_return(("sliding_get_sb_range: the pointer is out of range\n"), 0);
+}
+
 /* Since the sliding get_sb only allows us to access one byte at a time, we
  * do the same as we did with get_sb with the memcpy since one memcpy is much
  * faster than numerous memcpys 1 byte at a time */
@@ -199,22 +223,15 @@ static void single_mcpy(rzip_control *control, unsigned char *buf, i64 offset, i
 
 static void sliding_mcpy(rzip_control *control, unsigned char *buf, i64 offset, i64 len)
 {
-	struct sliding_buffer *sb = &control->sb;
-	i64 i, offdiff = offset - sb->offset_low;
+	i64 n = 0;
 
-	/* See if we fit in the low buffer first and use the faster function
-	 * where possible */
-	if (offdiff >= 0 && offdiff < sb->size_low) {
-		i64 minlen = MIN(len, sb->size_low - offdiff);
+	while (n < len) {
+		uchar *srcbuf = sliding_get_sb(control, offset + n);
+		i64 m = MIN(sliding_get_sb_range(control, offset + n), len - n);
 
-		memcpy(buf, control->sb.buf_low + offdiff, minlen);
-		len -= minlen;
+		memcpy(buf + n, srcbuf, m);
+		n += m;
 	}
-
-	/* We have no choice but to go fine-grained if there's any len left
-	 * since we will be paging */
-	for (i = 0; i < len; i++)
-		memcpy(buf + i, sliding_get_sb(control, offset + i), 1);
 }
 
 /* All put_u8/u32/vchars go to stream 0 */
@@ -234,7 +251,7 @@ static inline bool put_u32(rzip_control *control, void *ss, uint32_t s)
 }
 
 /* Put a variable length of bytes dependant on how big the chunk is */
-static inline bool put_vchars(rzip_control *control, void *ss, i64 s, int length)
+static bool put_vchars(rzip_control *control, void *ss, i64 s, i64 length)
 {
 	s = htole64(s);
 	if (unlikely(write_stream(control, ss, 0, (uchar *)&s, length)))
@@ -251,7 +268,8 @@ static bool put_header(rzip_control *control, void *ss, uchar head, i64 len)
 	return true;
 }
 
-static bool put_match(rzip_control *control, struct rzip_state *st, i64 p, i64 offset, i64 len)
+static inline bool put_match(rzip_control *control, struct rzip_state *st,
+			     i64 p, i64 offset, i64 len)
 {
 	do {
 		i64 ofs;
@@ -275,7 +293,8 @@ static bool put_match(rzip_control *control, struct rzip_state *st, i64 p, i64 o
 }
 
 /* write some data to a stream mmap encoded. Return -1 on failure */
-static int write_sbstream(rzip_control *control, void *ss, int stream, i64 p, i64 len)
+static inline int write_sbstream(rzip_control *control, void *ss, int stream,
+				 i64 p, i64 len)
 {
 	struct stream_info *sinfo = ss;
 
@@ -332,7 +351,7 @@ static inline tag increase_mask(tag tag_mask)
 	return (tag_mask << 1) | 1;
 }
 
-static bool minimum_bitness(struct rzip_state *st, tag t)
+static inline bool minimum_bitness(struct rzip_state *st, tag t)
 {
 	tag better_than_min = increase_mask(st->minimum_tag_mask);
 
@@ -405,7 +424,7 @@ static void insert_hash(struct rzip_state *st, tag t, i64 offset)
 
 /* Eliminate one hash entry with minimum number of lower bits set.
    Returns tag requirement for any new entries. */
-static tag clean_one_from_hash(rzip_control *control, struct rzip_state *st)
+static inline tag clean_one_from_hash(rzip_control *control, struct rzip_state *st)
 {
 	struct hash_entry *he;
 	tag better_than_min;
@@ -458,8 +477,9 @@ static inline tag full_tag(rzip_control *control, struct rzip_state *st, i64 p)
 	return ret;
 }
 
-static inline i64 match_len(rzip_control *control, struct rzip_state *st, i64 p0, i64 op, i64 end,
-			    i64 *rev)
+static inline i64
+match_len(rzip_control *control, struct rzip_state *st, i64 p0, i64 op,
+	  i64 end, i64 *rev)
 {
 	uchar *(*csb)(rzip_control *, i64);
 	i64 p, len;
@@ -491,8 +511,9 @@ static inline i64 match_len(rzip_control *control, struct rzip_state *st, i64 p0
 	return len;
 }
 
-static i64 find_best_match(rzip_control *control, struct rzip_state *st, tag t, i64 p, i64 end,
-			   i64 *offset, i64 *reverse)
+static inline i64
+find_best_match(rzip_control *control, struct rzip_state *st, tag t, i64 p,
+		i64 end, i64 *offset, i64 *reverse)
 {
 	struct hash_entry *he;
 	i64 length = 0;
@@ -571,18 +592,19 @@ static void *cksumthread(void *data)
 	return NULL;
 }
 
-static void cksum_update(rzip_control *control)
+static inline void cksum_update(rzip_control *control)
 {
 	pthread_t thread;
 
 	create_pthread(control, &thread, NULL, cksumthread, control);
 }
 
-static bool hash_search(rzip_control *control, struct rzip_state *st, double pct_base, double pct_multiple)
+static inline bool hash_search(rzip_control *control, struct rzip_state *st,
+			       double pct_base, double pct_multiple)
 {
+	i64 cksum_limit = 0, p, end, cksum_chunks, cksum_remains, i;
 	struct sliding_buffer *sb = &control->sb;
 	int lastpct = 0, last_chunkpct = 0;
-	i64 cksum_limit = 0, p, end;
 	tag t = 0;
 	struct {
 		i64 p;
@@ -605,11 +627,10 @@ static bool hash_search(rzip_control *control, struct rzip_state *st, double pct
 		/* 66% full at max. */
 		st->hash_limit = (1 << st->hash_bits) / 3 * 2;
 		st->hash_table = calloc(sizeof(st->hash_table[0]), (1 << st->hash_bits));
+		if (unlikely(!st->hash_table))
+			fatal_return(("Failed to allocate hash table in hash_search\n"), false);
 	}
 
-	if (unlikely(!st->hash_table))
-		fatal_return(("Failed to allocate hash table in hash_search\n"), false);
-
 	st->minimum_tag_mask = tag_mask;
 	st->tag_clean_ptr = 0;
 	st->cksum = 0;
@@ -629,10 +650,9 @@ static bool hash_search(rzip_control *control, struct rzip_state *st, double pct
 	}
 
 	while (p < end) {
-		i64 reverse, mlen, offset = 0;
+		i64 reverse, mlen, offset;
 
-		p++;
-		sb->offset_search = p;
+		sb->offset_search = ++p;
 		if (unlikely(sb->offset_search > sb->offset_low + sb->size_low))
 			remap_low_sb(control, &control->sb);
 
@@ -661,6 +681,7 @@ static bool hash_search(rzip_control *control, struct rzip_state *st, double pct
 		if ((t & st->minimum_tag_mask) != st->minimum_tag_mask)
 			continue;
 
+		offset = 0;
 		mlen = find_best_match(control, st, t, p, end, &offset, &reverse);
 
 		/* Only insert occasionally into hash. */
@@ -717,15 +738,29 @@ static bool hash_search(rzip_control *control, struct rzip_state *st, double pct
 		put_literal(control, st, st->last_match, st->chunk_size);
 
 	if (st->chunk_size > cksum_limit) {
+		/* Compute checksum. If the entire chunk is longer than maxram,
+		 * do it "per-partes" */
 		lock_mutex(control, &control->cksumlock);
 		control->checksum.len = st->chunk_size - cksum_limit;
+		cksum_chunks = control->checksum.len / control->maxram;
+		cksum_remains = control->checksum.len % control->maxram;
+
 		control->checksum.buf = malloc(control->checksum.len);
 		if (unlikely(!control->checksum.buf))
-			fatal_return(("Failed to malloc ckbuf in hash_search\n"), false);
-		control->do_mcpy(control, control->checksum.buf, cksum_limit, control->checksum.len);
-		st->cksum = CrcUpdate(st->cksum, control->checksum.buf, control->checksum.len);
+			fatal_return(("Failed to malloc ckbuf in hash_search2\n"), false);
+
+		for (i = 0; i < cksum_chunks; i++) {
+			control->do_mcpy(control, control->checksum.buf, cksum_limit, control->maxram);
+			cksum_limit += control->maxram;
+			st->cksum = CrcUpdate(st->cksum, control->checksum.buf, control->maxram);
+			if (!NO_MD5)
+				md5_process_bytes(control->checksum.buf, control->maxram, &control->ctx);
+		}
+		/* Process end of the checksum buffer */
+		control->do_mcpy(control, control->checksum.buf, cksum_limit, cksum_remains);
+		st->cksum = CrcUpdate(st->cksum, control->checksum.buf, cksum_remains);
 		if (!NO_MD5)
-			md5_process_bytes(control->checksum.buf, control->checksum.len, &control->ctx);
+			md5_process_bytes(control->checksum.buf, cksum_remains, &control->ctx);
 		free(control->checksum.buf);
 		unlock_mutex(control, &control->cksumlock);
 	} else
@@ -739,7 +774,7 @@ static bool hash_search(rzip_control *control, struct rzip_state *st, double pct
 }
 
 
-static void init_hash_indexes(struct rzip_state *st)
+static inline void init_hash_indexes(struct rzip_state *st)
 {
 	int i;
 
@@ -753,15 +788,29 @@ static void init_hash_indexes(struct rzip_state *st)
 
 static inline void *fake_mremap(void *old_address, size_t old_size, size_t new_size, int flags __UNUSED__)
 {
-	munmap(old_address, old_size);
-	return mmap(old_address, new_size, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
+	if (new_size > old_size) {
+		fprintf(stderr, "fake_mremap: This should only be used to shrink things. I'm not bothering with this.\n");
+		exit(1);
+	} else {
+		/* new_size occupies N pages; old_size occupies M > N pages;
+		 we want to unmap the M - N pages at the end.
+		 note the idiom: ceiling(n/k) = (n+k-1) div k */
+		size_t kept_n = (new_size + PAGE_SIZE - 1) / PAGE_SIZE;
+		int ret = munmap(old_address + (kept_n * PAGE_SIZE), old_size - (kept_n * PAGE_SIZE));
+
+		if (ret < 0)
+			return MAP_FAILED;
+
+		return old_address;
+	}
 }
 
 /* stdin is not file backed so we have to emulate the mmap by mapping
  * anonymous ram and reading stdin into it. It means the maximum ram
  * we can use will be less but we will already have determined this in
  * rzip_chunk */
-static bool mmap_stdin(rzip_control *control, uchar *buf, struct rzip_state *st)
+static inline bool mmap_stdin(rzip_control *control, uchar *buf,
+			      struct rzip_state *st)
 {
 	i64 len = st->chunk_size;
 	uchar *offset_buf = buf;
@@ -799,7 +848,9 @@ static bool mmap_stdin(rzip_control *control, uchar *buf, struct rzip_state *st)
 	return true;
 }
 
-static bool init_sliding_mmap(rzip_control *control, struct rzip_state *st, int fd_in, i64 offset)
+static inline bool
+init_sliding_mmap(rzip_control *control, struct rzip_state *st, int fd_in,
+		  i64 offset)
 {
 	struct sliding_buffer *sb = &control->sb;
 
@@ -825,8 +876,9 @@ static bool init_sliding_mmap(rzip_control *control, struct rzip_state *st, int
 
 /* compress a chunk of an open file. Assumes that the file is able to
    be mmap'd and is seekable */
-static bool rzip_chunk(rzip_control *control, struct rzip_state *st, int fd_in, int fd_out, i64 offset,
-		       double pct_base, double pct_multiple)
+static inline bool
+rzip_chunk(rzip_control *control, struct rzip_state *st, int fd_in, int fd_out,
+	   i64 offset, double pct_base, double pct_multiple)
 {
 	struct sliding_buffer *sb = &control->sb;
 
@@ -880,6 +932,7 @@ bool rzip_fd(rzip_control *control, int fd_in, int fd_out)
 	double chunkmbs;
 	i64 free_space;
 
+	init_mutex(control, &control->control_lock);
 	if (!NO_MD5)
 		md5_init_ctx(&control->ctx);
 	init_mutex(control, &control->cksumlock);
