@@ -1,6 +1,6 @@
 /*
    Copyright (C) 2011 Serge Belyshev
-   Copyright (C) 2006-2012 Con Kolivas
+   Copyright (C) 2006-2013 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998 Andrew Tridgell
 
@@ -304,6 +304,7 @@ static int gzip_compress_buf(rzip_control *control, struct compress_thread *cthr
 
 static int lzma_compress_buf(rzip_control *control, struct compress_thread *cthread)
 {
+	unsigned char lzma_properties[5]; /* lzma properties, encoded */
 	int lzma_level, lzma_ret;
 	size_t prop_size = 5; /* return value for lzma_properties */
 	uchar *c_buf;
@@ -325,10 +326,10 @@ retry:
 	}
 
 	/* with LZMA SDK 4.63, we pass compression level and threads only
-	 * and receive properties in control->lzma_properties */
+	 * and receive properties in lzma_properties */
 
 	lzma_ret = LzmaCompress(c_buf, &dlen, cthread->s_buf,
-		(size_t)cthread->s_len, control->lzma_properties, &prop_size,
+		(size_t)cthread->s_len, lzma_properties, &prop_size,
 				lzma_level,
 				0, /* dict size. set default, choose by level */
 				-1, -1, -1, -1, /* lc, lp, pb, fb */
@@ -375,6 +376,18 @@ retry:
 		return 0;
 	}
 
+	/* Make sure multiple threads don't race on writing lzma_properties */
+	lock_mutex(control, &control->control_lock);
+	if (!control->lzma_prop_set) {
+		memcpy(control->lzma_properties, lzma_properties, 5);
+		control->lzma_prop_set = true;
+		/* Reset the magic written flag so we write it again if we
+		 * get lzma properties and haven't written them yet. */
+		if (TMP_OUTBUF)
+			control->magic_written = 0;
+	}
+	unlock_mutex(control, &control->control_lock);
+
 	cthread->c_len = dlen;
 	free(cthread->s_buf);
 	cthread->s_buf = c_buf;
@@ -1325,9 +1338,13 @@ retry:
 		int j;
 
 		if (TMP_OUTBUF) {
+			lock_mutex(control, &control->control_lock);
 			if (!control->magic_written)
 				write_magic(control);
-			if (unlikely(!flush_tmpoutbuf(control))) goto error;
+			unlock_mutex(control, &control->control_lock);
+
+			if (unlikely(!flush_tmpoutbuf(control)))
+				goto error;
 		}
 
 		print_maxverbose("Writing initial chunk bytes value %d at %lld\n",
@@ -1741,9 +1758,8 @@ int close_stream_out(rzip_control *control, void *ss)
 	int i;
 
 	for (i = 0; i < sinfo->num_streams; i++) {
-		if (sinfo->s[i].buflen)
-			if (unlikely(!clear_buffer(control, sinfo, i, 0)))
-				return -1;
+		if (unlikely(!clear_buffer(control, sinfo, i, 0)))
+			return -1;
 	}
 
 	if (ENCRYPT) {
